<?php
/**
 * @link https://github.com/zhuravljov/yii2-queue
 * @copyright Copyright (c) 2017 Roman Zhuravlev
 * @license http://opensource.org/licenses/BSD-3-Clause
 */

namespace zhuravljov\yii\queue\redis;

use yii\base\InvalidParamException;
use yii\base\NotSupportedException;
use yii\di\Instance;
use yii\redis\Connection;
use zhuravljov\yii\queue\cli\Queue as CliQueue;
use zhuravljov\yii\queue\cli\Signal;

/**
 * Redis Queue
 *
 * @author Roman Zhuravlev <zhuravljov@gmail.com>
 */
class Queue extends CliQueue
{
    /**
     * @var Connection|array|string
     */
    public $redis = 'redis';
    /**
     * @var string
     */
    public $channel = 'queue';

    /**
     * @var string command class name
     */
    public $commandClass = Command::class;

    /**
     * @inheritdoc
     */
    public function init()
    {
        parent::init();
        $this->redis = Instance::ensure($this->redis, Connection::class);
    }

    /**
     * Runs all jobs from redis-queue.
     */
    public function run()
    {
        $this->openWorker();
        while (($payload = $this->reserve(0)) !== null) {
            list($id, $message, $ttr, $attempt) = $payload;
            if ($this->handleMessage($id, $message, $ttr, $attempt)) {
                $this->delete($id);
            }
        }
        $this->closeWorker();
    }

    /**
     * Listens redis-queue and runs new jobs.
     *
     * @param int $wait timeout
     */
    public function listen($wait)
    {
        $this->openWorker();
        while (!Signal::isExit()) {
            if (($payload = $this->reserve($wait)) !== null) {
                list($id, $message, $ttr, $attempt) = $payload;
                if ($this->handleMessage($id, $message, $ttr, $attempt)) {
                    $this->delete($id);
                }
            }
        }
        $this->closeWorker();
    }

    /**
     * @param int $wait timeout
     * @return array|null payload
     */
    protected function reserve($wait)
    {
        // Move delayed messages into waiting
        if ($this->now < time()) {
            $this->now = time();
            $this->moveExpired("$this->channel.delayed", $this->now);
            $this->moveExpired("$this->channel.reserved", $this->now);
        }

        // Find a new waiting message
        $id = null;
        if (!$wait) {
            $id = $this->redis->rpop("$this->channel.waiting");
        } elseif ($result = $this->redis->brpop("$this->channel.waiting", $wait)) {
            $id = $result[1];
        }
        if (!$id) {
            return null;
        }

        $payload = $this->redis->hget("$this->channel.messages", $id);
        list($ttr, $message) = explode(';', $payload, 2);
        $this->redis->zadd("$this->channel.reserved", time() + $ttr, $id);
        $attempt = $this->redis->hincrby("$this->channel.attempts", $id, 1);

        return [$id, $message, $ttr, $attempt];
    }

    private $now = 0;

    /**
     * @param string $from
     * @param int $time
     */
    protected function moveExpired($from, $time)
    {
        if ($expired = $this->redis->zrevrangebyscore($from, $time, '-inf')) {
            $this->redis->zremrangebyscore($from, '-inf', $time);
            foreach ($expired as $id) {
                $this->redis->rpush("$this->channel.waiting", $id);
            }
        }
    }

    /**
     * Deletes message by ID
     *
     * @param int $id of a message
     */
    protected function delete($id)
    {
        $this->redis->zrem("$this->channel.reserved", $id);
        $this->redis->hdel("$this->channel.attempts", $id);
        $this->redis->hdel("$this->channel.messages", $id);
    }

    /**
     * @inheritdoc
     */
    protected function pushMessage($message, $ttr, $delay, $priority)
    {
        if ($priority !== null) {
            throw new NotSupportedException('Job priority is not supported in the driver.');
        }

        $id = $this->redis->incr("$this->channel.message_id");
        $this->redis->hset("$this->channel.messages", $id, "$ttr;$message");
        if (!$delay) {
            $this->redis->lpush("$this->channel.waiting", $id);
        } else {
            $this->redis->zadd("$this->channel.delayed", time() + $delay, $id);
        }

        return $id;
    }

    protected function openWorker()
    {
        $id = $this->redis->incr("$this->channel.worker_id");
        $this->redis->clientSetname("$this->channel.worker.$id");
    }

    protected function closeWorker()
    {
        $this->redis->clientSetname('');
    }

    /**
     * @inheritdoc
     */
    protected function status($id)
    {
        if (!is_numeric($id) || $id <= 0) {
            throw new InvalidParamException("Unknown messages ID: $id.");
        }

        if ($this->redis->hexists("$this->channel.attempts", $id)) {
            return self::STATUS_RESERVED;
        } elseif ($this->redis->hexists("$this->channel.messages", $id)) {
            return self::STATUS_WAITING;
        } else {
            return self::STATUS_DONE;
        }
    }
}